const char snapshot_daemon_lua[] =
"-- snapshot_daemon.lua (internal file)\n"
"\n"
"do\n"
"    local log = require 'log'\n"
"    local fiber = require 'fiber'\n"
"    local fio = require 'fio'\n"
"    local yaml = require 'yaml'\n"
"    local errno = require 'errno'\n"
"    local digest = require 'digest'\n"
"    local pickle = require 'pickle'\n"
"\n"
"    local PREFIX = 'snapshot_daemon'\n"
"\n"
"    local daemon = {\n"
"        snapshot_period = 0;\n"
"        snapshot_period_bias = 0;\n"
"        snapshot_count = 6;\n"
"    }\n"
"\n"
"    local function sprintf(fmt, ...) return string.format(fmt, ...) end\n"
"\n"
"    -- create snapshot, return true if no errors\n"
"    local function snapshot()\n"
"        log.info(\"making snapshot...\")\n"
"        local s, e = pcall(function() box.snapshot() end)\n"
"        if s then\n"
"            return true\n"
"        end\n"
"        -- don't complain log if snapshot is already exists\n"
"        if errno() == errno.EEXIST then\n"
"            return false\n"
"        end\n"
"        log.error(\"error while creating snapshot: %s\", e)\n"
"        return false\n"
"    end\n"
"\n"
"    -- create snapshot by several options\n"
"    local function make_snapshot(last_snap)\n"
"\n"
"        if daemon.snapshot_period == nil then\n"
"            return false\n"
"        end\n"
"\n"
"        if not(daemon.snapshot_period > 0) then\n"
"            return false\n"
"        end\n"
"\n"
"\n"
"        if last_snap == nil then\n"
"            return snapshot()\n"
"        end\n"
"\n"
"        local vclock = box.info.vclock\n"
"        local lsn = 0\n"
"        for i, vlsn in pairs(vclock) do\n"
"            lsn = lsn + vlsn\n"
"        end\n"
"\n"
"        local snap_name = sprintf('%020d.snap', tonumber(lsn))\n"
"        if fio.basename(last_snap) == snap_name then\n"
"            if daemon.last_snap_name ~= snap_name then\n"
"                daemon.last_snap_name = snap_name\n"
"                log.debug('snapshot file %s already exists', last_snap)\n"
"            end\n"
"            return false\n"
"        end\n"
"        daemon.last_snap_name = snap_name\n"
"\n"
"        local snstat = fio.stat(last_snap)\n"
"        if snstat == nil then\n"
"            log.error(\"can't stat %s: %s\", last_snap, errno.strerror())\n"
"            return false\n"
"        end\n"
"        if snstat.mtime <= fiber.time() + daemon.snapshot_period then\n"
"            return snapshot()\n"
"        end\n"
"    end\n"
"\n"
"    -- check filesystem and current time\n"
"    local function process(self)\n"
"        local snaps = fio.glob(fio.pathjoin(box.cfg.snap_dir, '*.snap'))\n"
"\n"
"        if snaps == nil then\n"
"            log.error(\"can't read snap_dir %s: %s\", box.cfg.snap_dir,\n"
"                      errno.strerror())\n"
"            return\n"
"        end\n"
"\n"
"        if not make_snapshot(snaps[#snaps]) then\n"
"            return\n"
"        end\n"
"\n"
"        -- cleanup code\n"
"        if daemon.snapshot_count == nil then\n"
"            return\n"
"        end\n"
"\n"
"        if not (self.snapshot_count > 0) then\n"
"            return\n"
"        end\n"
"\n"
"\n"
"        -- reload snap list after snapshot\n"
"        snaps = fio.glob(fio.pathjoin(box.cfg.snap_dir, '*.snap'))\n"
"        local xlogs = fio.glob(fio.pathjoin(box.cfg.wal_dir, '*.xlog'))\n"
"        if xlogs == nil then\n"
"            log.error(\"can't read wal_dir %s: %s\", box.cfg.wal_dir,\n"
"                      errno.strerror())\n"
"            return\n"
"        end\n"
"\n"
"        while #snaps > self.snapshot_count do\n"
"            local rm = snaps[1]\n"
"            table.remove(snaps, 1)\n"
"\n"
"            log.info(\"removing old snapshot %s\", rm)\n"
"            if not fio.unlink(rm) then\n"
"                log.error(\"error while removing %s: %s\",\n"
"                          rm, errno.strerror())\n"
"                return\n"
"            end\n"
"        end\n"
"\n"
"\n"
"        local snapno = fio.basename(snaps[1], '.snap')\n"
"\n"
"        while #xlogs > 0 do\n"
"            if #xlogs < 2 then\n"
"                break\n"
"            end\n"
"\n"
"            if fio.basename(xlogs[1], '.xlog') > snapno then\n"
"                break\n"
"            end\n"
"\n"
"            if fio.basename(xlogs[2], '.xlog') > snapno then\n"
"                break\n"
"            end\n"
"\n"
"\n"
"            local rm = xlogs[1]\n"
"            table.remove(xlogs, 1)\n"
"            log.info(\"removing old xlog %s\", rm)\n"
"\n"
"            if not fio.unlink(rm) then\n"
"                log.error(\"error while removing %s: %s\",\n"
"                          rm, errno.strerror())\n"
"                return\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"\n"
"    local function next_snap_interval(self)\n"
"\n"
"        -- don't do anything in hot_standby mode\n"
"        if box.info.status ~= 'running' or\n"
"            self.snapshot_period == nil or\n"
"            self.snapshot_period <= 0 then\n"
"            return nil\n"
"        end\n"
"\n"
"        return self.snapshot_period -\n"
"            (fiber.time() + self.snapshot_period_bias) % self.snapshot_period\n"
"    end\n"
"\n"
"    local function daemon_fiber(self)\n"
"        fiber.name(PREFIX)\n"
"        log.info(\"started\")\n"
"        while true do\n"
"            local interval = next_snap_interval(self)\n"
"            if interval == nil then\n"
"                break\n"
"            end\n"
"            fiber.sleep(interval)\n"
"            if self.reloaded ~= true then\n"
"                local s, e = pcall(process, self)\n"
"\n"
"                if not s then\n"
"                    log.error(e)\n"
"                end\n"
"            else\n"
"                self.reloaded = false\n"
"                log.info(\"reloaded\")\n"
"            end\n"
"        end\n"
"        log.info(\"stopped\")\n"
"        self.fiber = nil\n"
"    end\n"
"\n"
"    local function reload(self)\n"
"        if self.snapshot_period > 0 and self.fiber == nil then\n"
"            self.fiber = fiber.create(daemon_fiber, self)\n"
"        elseif self.fiber ~= nil then\n"
"            -- wake up daemon\n"
"            self.reloaded = true\n"
"            fiber.wakeup(self.fiber)\n"
"        end\n"
"    end\n"
"\n"
"    setmetatable(daemon, {\n"
"        __index = {\n"
"            set_snapshot_period = function()\n"
"                daemon.snapshot_period = box.cfg.snapshot_period\n"
"                if daemon.snapshot_period > 0 then\n"
"                    local rnd = pickle.unpack('i', digest.urandom(4))\n"
"                    daemon.snapshot_period_bias = rnd % daemon.snapshot_period\n"
"                else\n"
"                    daemon.snapshot_period_bias = 0\n"
"                end\n"
"                reload(daemon)\n"
"                return\n"
"            end,\n"
"\n"
"            set_snapshot_count = function()\n"
"                if math.floor(box.cfg.snapshot_count) ~= box.cfg.snapshot_count then\n"
"                    box.error(box.error.CFG, \"snapshot_count\",\n"
"                             \"must be an integer\")\n"
"                end\n"
"                daemon.snapshot_count = box.cfg.snapshot_count\n"
"                reload(daemon)\n"
"            end,\n"
"\n"
"            next_snap_interval = next_snap_interval\n"
"        }\n"
"    })\n"
"\n"
"    if box.internal == nil then\n"
"        box.internal = { [PREFIX] = daemon }\n"
"    else\n"
"        box.internal[PREFIX] = daemon\n"
"    end\n"
"end\n"
""
;
